/* Developed for Lenovo Educational Board Pass 3
   © Lenovo Created by Joshua Grosser
   v. 2.0 */

#include <WiFi.h>
#include <WiFiMulti.h>
#include <WiFiClientSecure.h>
#include <WebSocketsClient_Generic.h>
#include <Preferences.h>

Preferences preferences;
WiFiMulti WiFiMulti;
WebSocketsClient  webSocket;

// change this to the ip adress printed to the serial on the server
#define WS_SERVER "192.168.1.148"
#define WS_PORT 8080
#define _WEBSOCKETS_LOGLEVEL_ 2

// define the LED pin
#define PIN_OUT 15
// define the unit length in ms
#define UNIT_LENGTH 250
// define the button pin to input morse code
#define BUTTON 4
// define the button pin to read morse code
const int buttonTwo = 5;
// define the buton to listen
const int buttonThree = 6;
// determines whether to output to the LED
bool complete = false; 
// message in letters
String finalMsg = ""; 
// used for translation to and from morse code
static String codex = "ABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890 .,?!:;()@&"; 
String ssid = "";
String password = "";

// array where the translations to morse code are stored
static String dots_and_dashes[] = {
  ".-",
  "-...",
  "-.-.",
  "-..",
  ".",
  "..-.",
  "--.",
  "....",
  "..",
  ".---",
  "-.-",
  ".-..",
  "--",
  "-.",
  "---",
  ".--.",
  "--.-",
  ".-.",
  "...",
  "-",
  "..-",
  "...-",
  ".--",
  "-..-",
  "-.--",
  "--..",
  ".----",
  "..---",
  "...--",
  "....-",
  ".....",
  "-....",
  "--...",
  "---..",
  "----.",
  "-----",
  "       ",
  ".-.-.-",
  "--..--",
  "..--..",
  "-.-...",
  "---...",
  "-.-.-.",
  "-.--.",
  "-.--.-",
  ".--.-.",
  ".-...",
};

// get Wifi info
void getWifi()
{
  preferences.begin("credentials", false);
  ssid = preferences.getString("ssid", ""); 
  password = preferences.getString("password", "");
  if (ssid == "" || password == ""){
    Serial.println("No ssid or password saved.");
  }
  else {
    Serial.println("Successfully read WiFi information.");
  }
}

bool alreadyConnected = false;
// receives an event from websocket and processes whether event is text, disconnected, or connnected
void webSocketEvent(const WStype_t& type, uint8_t * payload, const size_t& length)
{
  switch (type)
  {
    case WStype_DISCONNECTED:
    {
      if (alreadyConnected)
      {
        Serial.println("Disconnected!");
        alreadyConnected = false;
      }
      break;
    }
    case WStype_CONNECTED:
      {
        alreadyConnected = true;
        Serial.println("Connected!");
        break;
      }
    case WStype_TEXT:
    {
      Serial.println();
      Serial.printf("Text Recieved: %s\n", payload);
      String str = (char*)payload;
      outputToLED(str);
      break;
    }
    default:
    {
      break;
    }
  }
}

void setup()
{
  Serial.begin(115200);
  getWifi();
  // sets pins to appropriate mode
  pinMode(BUTTON, INPUT);
  pinMode(buttonTwo, INPUT);
  pinMode(buttonThree, INPUT);
  pinMode(PIN_OUT, OUTPUT);
  digitalWrite(PIN_OUT, LOW);
  pinMode(16, OUTPUT);
  digitalWrite(16, LOW);
  delay(200);
  Serial.setDebugOutput(true);
  // connect to Wi-Fi
  WiFiMulti.addAP(ssid.c_str(), password.c_str());
  while (WiFiMulti.run() != WL_CONNECTED)
  {
    Serial.print("."); 
    delay(100);
  }
  Serial.println();
  // Client address
  Serial.print("WebSockets Client started @ IP address: ");
  Serial.println(WiFi.localIP());
  // server address and port
  Serial.print("Connecting to WebSockets Server @ ");
  Serial.println(WS_SERVER);
  // begin web socket
  webSocket.begin(WS_SERVER, WS_PORT, "/");
  webSocket.onEvent(webSocketEvent);
  // try 5 seconds if no connection
  webSocket.setReconnectInterval(5000);
  // connected to web socket server
  Serial.print("Connected to WebSockets Server @ IP address: ");
  Serial.println(WS_SERVER);
}

void loop(){
  // need to be in this mode to receive a message
  while(digitalRead(buttonThree) != HIGH)
  {
    webSocket.loop();
  }
  // press button two after inputting a character to send message
  while (digitalRead(buttonTwo) != HIGH)
  {
    String myMsg = GetLetter();
    Translate(myMsg);
  }
  webSocket.sendTXT(finalMsg);
  finalMsg = "";
  Serial.println();
}

// outputs string to LEDs in morse code
void outputToLED(String message)
{
  String morseWord = encode(message);
  for(int i = 0; i <= morseWord.length(); i++)
  {
    switch(morseWord[i])
    {
      case '.': // dit
        digitalWrite(PIN_OUT, HIGH);
        delay(UNIT_LENGTH);
        digitalWrite(PIN_OUT, LOW);
        delay(UNIT_LENGTH);
        break;

      case '-': // dah
        digitalWrite(PIN_OUT, HIGH);
        delay(UNIT_LENGTH * 3);
        digitalWrite(PIN_OUT, LOW);
        delay(UNIT_LENGTH);
        break;

      case ' ': // gap
        delay(UNIT_LENGTH);
    }
  }
  digitalWrite(16, HIGH);
  delay(UNIT_LENGTH);
  digitalWrite(16, LOW);
  delay(UNIT_LENGTH * 3);
}

// reads input from button to morse code
String GetLetter()
{
  unsigned long START, STOP, TIME;
  String myMsg = "";
  while(true)
  {
    while(digitalRead(BUTTON) == LOW)
    {
      // do nothing until button is pressed
    }
    digitalWrite(PIN_OUT, HIGH);
    START = millis();
    while(digitalRead(BUTTON) == HIGH)
    {
      // do not register input until button is released
    }
    STOP = millis();
    digitalWrite(PIN_OUT, LOW);
    TIME = STOP - START;
    if (TIME > 25)
    {
      myMsg += Dot_or_Dash(TIME);
    }
    while((millis() - STOP) < 750)
    {
      delay(10);
      if (digitalRead(BUTTON) == HIGH)
      {
        break;
      }
    }
    if ((millis() - STOP) >= 750)
    {
      return myMsg;
    }
  }
}

// determine if it is a dot or dash
char Dot_or_Dash (unsigned long TIME)
{
  char DoD;
  if (TIME < 500)
  {
    DoD = '.';
  }
  else if (TIME >= 500 && TIME < 5000)
  {
    DoD = '-';
  }
  else
  {
    DoD = ' ';
  }
  return DoD;
}

// translates morse code to letters
void Translate (String myMsg)
{
  for (int i = 0; i < 47; i++)
  {
    if (myMsg == dots_and_dashes[i])
    {
      Serial.print(codex[i]);
      finalMsg += codex[i];
      return ;
    }
  }
  Serial.print("ERROR, I cannot understand you!\n");
  return ;
}

// translates letters to morse code
String encode(String string)
{
  size_t i, j;
  String morseWord = "";
  for( i = 0; string[i]; i++ )
  {
    for( j = 0; j < 47; j++ )
    {
      if( string[i] == codex[j] )
      {
        morseWord += dots_and_dashes[j];
        break;
      }
    }
    morseWord += " "; // Add tailingspace to separate the chars
  }
  return morseWord;
}
