/* Developed for Lenovo Educational Board Pass 3
   © Lenovo Created by Joshua Grosser
   v. 2.0 */

// libraries to be used for OLED display and time
#include <Arduino.h>
#include <WiFi.h>
#include <time.h>
#include <SPI.h>
#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#include <Preferences.h>

Preferences preferences;
Adafruit_SSD1306 display = Adafruit_SSD1306(128, 64, &Wire, -1); 

String ssid = "";
String password = "";

int GMTOffset = 0;  // GMT Offset counts up from GMT time to your current time
bool setOffset = false; // whether offset is displayed or not

// get Wifi info
void getWifi()
{
  preferences.begin("credentials", false);
  ssid = preferences.getString("ssid", ""); 
  password = preferences.getString("password", "");
  if (ssid == "" || password == ""){
    Serial.println("No ssid or password saved.");
  }
  else {
    Serial.println("Successfully read WiFi information.");
  }
}
 
void setup() {
  // set baud rate
  Serial.begin(115200); 
  // initialize display
  display.begin(SSD1306_SWITCHCAPVCC, 0x3C);
  delay(200);
  display.clearDisplay();
  display.setTextSize(1); 
  display.setCursor(0,0);
  display.setTextColor(WHITE);
  // connect to WiFi
  getWifi();
  WiFi.begin(ssid.c_str(), password.c_str());
  Serial.println("Connecting...");
  while (WiFi.status() != WL_CONNECTED) 
  {
    delay(1000);
    Serial.print(".");
  }
  Serial.println();
  Serial.println("Connected to Wi-Fi!");
}

void loop() {
  while (!setOffset)
  { // indicates whether you have put in your current offset in relation to GMT time
    while(Serial.available() > 0)
    {
      int hours = Serial.parseInt();
      int minutes = Serial.parseInt();
      // A newline character indicates end of input
      if(Serial.read() == '\n')
      {
        // converts to current time
        GMTOffset = (hours * 3600) + (minutes * 60);
        configTime(GMTOffset, 0, "pool.ntp.org","time.nist.gov"); // sets time
        setOffset = true;
      }
    }
  }
  time_t rawtime = time(nullptr);
  struct tm* timeinfo = localtime(&rawtime);

  int radius = 35;
  display.drawCircle(display.width()/2, display.height()/2, 2, WHITE);
  // draws the clock face
  for( int i = 0; i < 360; i= i + 30 ){
  float angle = i; 
  angle = (angle/57.29577951); 
  int x1 =(64+(sin(angle)*radius));
  int y1 =(32-(cos(angle)*radius));
  int x2 =(64+(sin(angle)*(radius-5)));
  int y2 =(32-(cos(angle)*(radius-5)));
  display.drawLine(x1,y1,x2,y2,WHITE);
  }
  
  // draws the second hand
  float angle = timeinfo->tm_sec*6;
  angle = (angle/57.29577951); 
  int x2 = (64+(sin(angle)*(radius)));
  int y2 = (32-(cos(angle)*(radius)));
  display.drawLine(64,32,x2,y2,WHITE);
  
  // draws the minute hand
  angle = timeinfo->tm_min * 6;
  angle = (angle/57.29577951); 
  x2 =(64+(sin(angle)*(radius-3)));
  y2 =(32-(cos(angle)*(radius-3)));
  display.drawLine(64,32,x2,y2,WHITE);
  
  // draws the hour hand
  angle = timeinfo->tm_hour * 30 + int((timeinfo->tm_min / 12) * 6);
  angle = (angle/57.29577951); 
  x2 =(64+(sin(angle)*(radius-11)));
  y2 =(32-(cos(angle)*(radius-11)));
  display.drawLine(64,32,x2,y2,WHITE);

  // shows info to display then resets
  display.display();
  delay(100);
  display.clearDisplay(); 
}
