/* Developed for Lenovo Educational Board Pass 3
   © Lenovo Created by Grason Humphrey & Modified by Joshua Grosser
   v. 2.0 */

#include <Wire.h>
#include "SSD1306Wire.h"

SSD1306Wire display(0x3c);
#define WIRE Wire

// Define game constants

// Width and height of screen
const int SCREEN_WIDTH = 128;
const int SCREEN_HEIGHT = 64;
// Game loop timer, lower value means faster movement
const int TIMER_ABS_MAX = 500;
const int TIMER_ABS_MIN = 150;
// How fast does the difficulty increase
const int TIMER_STEP = 10;
int timerMax = TIMER_ABS_MAX;
// Location in array for X and Y coordinates
const int X_ARRAY = 0;
const int Y_ARRAY = 1;
// Size of individual piece block
const int BLOCK_SIZE = 6;
// Max number of blocks that can fit on the screen
const int MAX_BLOCKS = 210;
const int MAX_BLOCKS_X = 10;
const int MAX_BLOCKS_Y = 21;
const int SHAPE_SIZE = 4;
const int NUM_BLOCK_TYPES = 7;

int randCount = 0;
int randBlocks[7] = {0, 1, 2, 3, 4, 5, 6};

int score = 0;

int timerCount = timerMax;

int dropCount = timerMax;
int dropMax = timerMax;

bool lReady = true;
bool rReady = true;
bool uReady = true;
bool dReady = true;

// Number of blocks currently on the screen
int numBlocks = 0;
int lastBlock = -1;

// Direction buttons for rotation control
int left = 5;
int down = 7;
int up = 4;
int right = 6;

struct playerBlock {
  bool shape[SHAPE_SIZE][SHAPE_SIZE];
  bool alive;
  int x;
  int y;
};

struct block {
  bool alive;
  int x;
  int y;
};

struct playerBlock myBlock;

bool otherBlocks[MAX_BLOCKS_Y][MAX_BLOCKS_X];

// shapes that randomly spawn
bool Square[SHAPE_SIZE][SHAPE_SIZE] = {
  {0, 0, 0, 0},
  {0, 1, 1, 0},
  {0, 1, 1, 0},
  {0, 0, 0, 0}
};

bool TShape[SHAPE_SIZE][SHAPE_SIZE] = {
  {0, 0, 0, 0},
  {0, 1, 0, 0},
  {1, 1, 1, 0},
  {0, 0, 0, 0}
};

bool LWiggle[SHAPE_SIZE][SHAPE_SIZE] = {
  {0, 0, 0, 0},
  {1, 1, 0, 0},
  {0, 1, 1, 0},
  {0, 0, 0, 0}
};

bool RWiggle[SHAPE_SIZE][SHAPE_SIZE] = {
  {0, 0, 0, 0},
  {0, 1, 1, 0},
  {1, 1, 0, 0},
  {0, 0, 0, 0}
};

bool LShape[SHAPE_SIZE][SHAPE_SIZE] = {
  {0, 0, 0, 0},
  {1, 0, 0, 0},
  {1, 1, 1, 0},
  {0, 0, 0, 0}
};

bool Bar[SHAPE_SIZE][SHAPE_SIZE] = {
  {0, 0, 0, 0},
  {1, 1, 1, 1},
  {0, 0, 0, 0},
  {0, 0, 0, 0}
};

bool RevLShape[SHAPE_SIZE][SHAPE_SIZE] = {
  {0, 0, 0, 0},
  {0, 0, 1, 0},
  {1, 1, 1, 0},
  {0, 0, 0, 0}
};

void setup() {
  // Start serial communication for debug
  Serial.begin(115200);
  // Start I2C communication with OLED display
  Wire.begin(8, 9);

  RandomShapeCopy(&myBlock);
  myBlock.alive = true;
  myBlock.x = 4;
  myBlock.y = 0;

  // Assign input buttons
  pinMode(left, INPUT);
  pinMode(down, INPUT);
  pinMode(up, INPUT);
  pinMode(right, INPUT);

  pinMode(45, INPUT);

  // I2C display
  display.init();
  display.setContrast(255);

}

void loop() {
  // Flag keeps track of if game has ended
  bool gameOver = false;

  // Timer to control game updating speed
  ++timerCount;
  ++dropCount;

  if (dropCount > dropMax) {
    myBlock.y++;
    dropCount = 0;
    bool landed = false;

    // Check if player shape has collided with floor
    for (int i = 0; i < SHAPE_SIZE; i++) {
      for (int j = 0; j < SHAPE_SIZE; j++) {
        if (myBlock.shape[i][j] && (myBlock.y + i >= MAX_BLOCKS_Y) && !landed) {
          landed = true;
        }
        // Check if player has collided with another block
        if (otherBlocks[myBlock.y + i - 1][myBlock.x + j - 1] && myBlock.shape[i][j] && !landed) {
          landed = true;
        }
      }
    }


    // If player shape has landed, convert to blocks and reset player shape
    if (landed) {
      myBlock.y--;
      for (int i = 0; i < SHAPE_SIZE; i++) {
        for (int j = 0; j < SHAPE_SIZE; j++) {
          if (myBlock.shape[i][j]) {
            otherBlocks[myBlock.y + i - 1][myBlock.x + j - 1] = true;
          }
        }
      }
      ClearFinishedLines();
      DrawScreen();

      // Check if player has lost
      for (int i = 0; i < MAX_BLOCKS_X; i++) {
        if (otherBlocks[0][i]) {
          display.clear();
          display.setTextAlignment(TEXT_ALIGN_CENTER);
          display.drawString(SCREEN_WIDTH / 2, 16, "Game Over");
          display.drawString(SCREEN_WIDTH / 2, 32, String("Score: ") + String(score));
          display.display();
          while (true) {
            delay(10);
          }
        }
      }

      // Reset block and increase speed
      myBlock.x = 4;
      myBlock.y = 0;
      RandomShapeCopy(&myBlock);
      if(timerMax > TIMER_ABS_MIN){
        timerMax -= TIMER_STEP;
      }
      if(dropMax > TIMER_ABS_MIN){
        dropMax -= TIMER_STEP;
      }
    }

  } else if (timerCount > timerMax) {
    // Time to update game
    DrawScreen();
    timerCount = 0;
  }

  // moves the block to the left
  if (digitalRead(left)) {
    if (lReady) {
      bool blocked = false;
      for (int i = 0; i < SHAPE_SIZE; i++) {
        for (int j = 0; j < SHAPE_SIZE; j++) {
          if (myBlock.shape[i][j] && ((myBlock.x + j - 2 < 0) || otherBlocks[myBlock.y + i - 1][myBlock.x + j - 2])) {
            blocked = true;
          }
        }
      }
      if (!blocked) {
        myBlock.x--;
      }
      timerCount = timerMax;
    }
    lReady = false;
  } else {
    lReady = true;
  }
  // moves the block to the right 
  if (digitalRead(right)) {
    if (rReady) {
      bool blocked = false;
      for (int i = 0; i < SHAPE_SIZE; i++) {
        for (int j = 0; j < SHAPE_SIZE; j++) {
          if (myBlock.shape[i][j] && ((myBlock.x + j >= MAX_BLOCKS_X) || otherBlocks[myBlock.y + i - 1][myBlock.x + j])) {
            blocked = true;
          }
        }
      }
      if (!blocked) {
        myBlock.x++;
      }
      timerCount = timerMax;
    }
    rReady = false;
  } else {
    rReady = true;
  }
  // rotates the block
  if (digitalRead(up)) {
    if (uReady) {
      timerCount = timerMax;
      RotatePlayerShape();
    }
    uReady = false;
  } else {
    uReady = true;
  }
  // moves the block down until it lands
  if (digitalRead(down)) {
    //if (dReady) {
      bool blocked = false;
      for (int i = 0; i < SHAPE_SIZE; i++) {
        for (int j = 0; j < SHAPE_SIZE; j++) {
          if (myBlock.shape[i][j] && ((myBlock.y + i + 1 >= MAX_BLOCKS_Y) || otherBlocks[myBlock.y + i][myBlock.x + j - 1])) {
            blocked = true;
          }
        }
      }
      if (!blocked) {
        myBlock.y++;
      }
      timerCount = timerMax;
    dReady = false;
  } else {
    dReady = true;
  }

  delay(1);
}

void ClearFinishedLines() {
  // clears a line if it is full
  int numLinesFinished = 0;
  bool lineFinished = true;

  for (int i = MAX_BLOCKS_Y - 1; i >= 0; i--) {
    if (i < (MAX_BLOCKS_Y - 1) && lineFinished) {
      ++numLinesFinished;
    }
    lineFinished = true;
    for (int j = 0; j < MAX_BLOCKS_X; j++) {
      if (numLinesFinished > 0) {
        otherBlocks[i + numLinesFinished][j] = otherBlocks[i][j];
      }
      if (!otherBlocks[i][j]) {
        lineFinished = false;
      }
    }
  }
  // Incrememt score, clearing more lines at once gives higher score
  if (numLinesFinished == 1) {
    score += 40;
  } else if (numLinesFinished == 2) {
    score += 100;
  } else if (numLinesFinished == 3) {
    score += 300;
  } else if (numLinesFinished >= 4) {
    score += 1200;
  }
}

void DrawScreen() {
  // Display
  display.clear();

  // Draw shape being controlled
  for (int i = 0; i < SHAPE_SIZE; i++) {
    for (int j = 0; j < SHAPE_SIZE; j++) {
      if (myBlock.shape[j][i]) {
        display.drawRect((myBlock.y + j) * BLOCK_SIZE + 1, SCREEN_HEIGHT - ((myBlock.x + i) * BLOCK_SIZE) - 2, BLOCK_SIZE, BLOCK_SIZE);
      }
    }
  }

  // Draw all blocks that have settled
  for (int i = 0; i < MAX_BLOCKS_Y; i++) {
    for (int j = 0; j < MAX_BLOCKS_X; j++) {
      if (otherBlocks[i][j]) {
        display.drawRect((i + 1) * BLOCK_SIZE + 1, SCREEN_HEIGHT - (j + 1) * BLOCK_SIZE - 2, BLOCK_SIZE, BLOCK_SIZE);
      }
    }
  }
  // Draw screen outline
  display.drawRect(0, 1, SCREEN_WIDTH, SCREEN_HEIGHT - 2);
  display.display();
}

void ShapeCopy(struct playerBlock *dest, bool shape[SHAPE_SIZE][SHAPE_SIZE]) {
  // copies a shape
  for (int i = 0; i < SHAPE_SIZE; i++) {
    for (int j = 0; j < SHAPE_SIZE; j++) {
      dest->shape[i][j] = shape[i][j];
    }
  }
}

void RandomShapeCopy(struct playerBlock *dest) {
  // copies a random shape
  int shape = random(NUM_BLOCK_TYPES - randCount);
  switch (randBlocks[shape]) {
    case 0:
      ShapeCopy(&myBlock, Square);
      break;
    case 1:
      ShapeCopy(&myBlock, TShape);
      break;
    case 2:
      ShapeCopy(&myBlock, LWiggle);
      break;
    case 3:
      ShapeCopy(&myBlock, RWiggle);
      break;
    case 4:
      ShapeCopy(&myBlock, LShape);
      break;
    case 5:
      ShapeCopy(&myBlock, Bar);
      break;
    case 6:
      ShapeCopy(&myBlock, RevLShape);
      break;
  }
  ++randCount;
  if (randCount < NUM_BLOCK_TYPES) {
    for (int i = shape; i < NUM_BLOCK_TYPES - randCount; i++) {
      randBlocks[i] = randBlocks[i + 1];
    }
  } else {
    for (int i = 0; i < NUM_BLOCK_TYPES; i++) {
      randBlocks[i] = i;
    }
    randCount = 0;
  }
}

void RotatePlayerShape() {
  bool tempArray[SHAPE_SIZE][SHAPE_SIZE];
  bool temp[SHAPE_SIZE][SHAPE_SIZE];
  // rotates the shape
  for (int i = 0; i < SHAPE_SIZE; i++) {
    for (int j = 0; j < SHAPE_SIZE; j++) {
      tempArray[SHAPE_SIZE - j - 1][i] = myBlock.shape[i][j];
      temp[i][j] = myBlock.shape[i][j];
    }
  }
  for (int i = 0; i < SHAPE_SIZE; i++) {
    for (int j = 0; j < SHAPE_SIZE; j++) {
      myBlock.shape[i][j] = tempArray[i][j];
      while (myBlock.shape[i][j] && (myBlock.x + j - 1 < 0) ) {
        myBlock.x++;
      }
      while (myBlock.shape[i][j] && (myBlock.x + j - 1 >= MAX_BLOCKS_X) ) {
        myBlock.x--;
      }
      while (myBlock.shape[i][j] && (myBlock.y + i - 1 >= MAX_BLOCKS_Y) ) {
        myBlock.y--;
      }
    }
  }
  // checks if shape can be rotated or if shape will clip into other shapes
  bool landed = false;
  // Check if player shape has collided with floor
  for (int i = 0; i < SHAPE_SIZE; i++) {
    for (int j = 0; j < SHAPE_SIZE; j++) {
      if (myBlock.shape[i][j] && (myBlock.y + i >= MAX_BLOCKS_Y) && !landed) {
        landed = true;
      }
      // Check if player has collided with another block
      if (otherBlocks[myBlock.y + i - 1][myBlock.x + j - 1] && myBlock.shape[i][j] && !landed) {
        landed = true;
      }
    }
  }
  if(landed)
  {
    for (int i = 0; i < SHAPE_SIZE; i++) {
      for (int j = 0; j < SHAPE_SIZE; j++) {
        myBlock.shape[i][j] = temp[i][j];
      }
    }
  }
}
